@@ -22,6 +22,8 @@ from parsley import makeGrammar
 import platform
 import subprocess
 import sys
+from packaging.version import parse as as_ver
+import distro
 
 
 debversion_grammar = """
@@ -45,7 +47,7 @@ rule = <name>:name selector?:selector version?:version ('\n'|comment) -> (
     name, selector or [], version or [])
 lowercase = ('a'|'b'|'c'|'d'|'e'|'f'|'g'|'h'|'i'|'j'|'k'|'l'|'m'|'n'|'o'|'p'
             |'q'|'r'|'s'|'t'|'u'|'v'|'w'|'x'|'y'|'z')
-name = letterOrDigit:start (letterOrDigit|'.'|'+'|'-'|'_'|'/')+:rest
+name = (letterOrDigit|'/'):start (letterOrDigit|'.'|'+'|'-'|'_'|'/')+:rest
 ws = ' '+
 profile = ('!'?:neg <(lowercase|digit|':'|'-'|'.')+>:name) -> (neg!='!', name)
 profiles = '(' (ws? profile)*:p ws? ')' -> p
@@ -58,13 +60,18 @@ comment = ws? '#' any* '\n' -> None
 any = ~'\n' anything
 blank = ws? '\n' -> None
 """
+PY2 = "base-py2"
+PY3 = "base-py3"
 
 
-def get_depends(filename=None):
+def get_depends(filename=None, default_blank=False):
     fd = get_depends_file(filename)
     if not fd:
-        return None
-    return Depends(fd.read())
+        if default_blank:
+            return Depends("")
+        else:
+            return None
+    return Depends(fd.read(), filename)
 
 
 def get_depends_file(filename=None):
@@ -124,15 +131,21 @@ class Depends(object):
         (True, True): True,
     }
 
-    def __init__(self, depends_string):
+    def __init__(self, depends_string, filename=None):
         """Construct a Depends instance.
 
         :param depends_string: The string description of the requirements that
             need to be satisfied. See the bindep README.rst for syntax for the
             requirements list.
+        :param filename: The string name of the file from which requirements
+            were loaded.
         """
+        if not depends_string.endswith('\n'):
+            # The parsley grammar expects each line to end with a newline
+            depends_string += '\n'
         parser = makeGrammar(grammar, {})(depends_string)
         self._rules = parser.rules()
+        self.filename = filename
 
     def _partition(self, rule):
         """Separate conditions into platform and user profiles.
@@ -232,10 +245,10 @@ class Depends(object):
 
         :param rules: A list of rules, as returned by active_rules.
         :param output_format: The format to print the output in. Currently
-        we support newline format which will print 1 package per line, and
-        csv format which prints a csv list.
+          we support newline format which will print 1 package per line, and
+          csv format which prints a csv list.
         :return: List of all required packages regardless of whether they are
-        missing.
+          missing.
         """
         packages_list = [rule[0] for rule in rules]
         if output_format == 'csv':
@@ -271,65 +284,148 @@ class Depends(object):
     def profiles(self):
         profiles = set()
         for rule in self._rules:
-            for _, selector in rule[1]:
-                profiles.add(selector)
+            # Partition rules, but keep only the user ones
+            _, user_profiles = self._partition(rule)
+            for profile in user_profiles:
+                # Flatten a series of AND conditionals in a user rule
+                if isinstance(profile, list):
+                    profiles.update([rule[1] for rule in profile])
+                elif isinstance(profile, tuple):
+                    profiles.add(profile[1])
         return sorted(profiles)
 
+    def codenamebits(self, distro_id, codename):
+        atoms = set()
+        codenamebits = codename.split()
+        for i in range(len(codenamebits)):
+            atoms.add("%s-%s" % (distro_id, "-".join(codenamebits[:i + 1])))
+        return atoms
+
+    def releasebits(self, distro_id, release):
+        atoms = set()
+        releasebits = release.split(".")
+        for i in range(len(releasebits)):
+            atoms.add("%s-%s" % (distro_id, ".".join(releasebits[:i + 1])))
+        return atoms
+
     def platform_profiles(self):
         if platform.system() == 'Darwin':
             atoms = set(['darwin'])
             # detect available macos package managers
             if os.system('which brew >/dev/null') == 0:
                 atoms.add('brew')
+                atoms.add(PY2)
                 self.platform = Brew()
             return ["platform:%s" % (atom,) for atom in sorted(atoms)]
-        try:
-            output = subprocess.check_output(
-                ["lsb_release", "-cirs"],
-                stderr=subprocess.STDOUT).decode(getpreferredencoding(False))
-        except OSError:
+        distro_id = distro.id()
+        if not distro_id:
             log = logging.getLogger(__name__)
-            log.error('Unable to execute lsb_release. Is it installed?')
-            raise
-        lsbinfo = output.lower().split()
+            log.error('Unable to determine distro ID. '
+                      'Does /etc/os-release exist or '
+                      'is lsb_release installed?')
+            raise Exception('Distro name not found')
         # NOTE(toabctl): distro can be more than one string (i.e. "SUSE LINUX")
-        codename = lsbinfo[len(lsbinfo) - 1:len(lsbinfo)][0]
-        release = lsbinfo[len(lsbinfo) - 2:len(lsbinfo) - 1][0]
+        codename = distro.codename().lower()
+        release = distro.version().lower()
+        release_version = as_ver(release)
         # NOTE(toabctl): space is a delimiter for bindep, so remove the spaces
-        distro = "".join(lsbinfo[0:len(lsbinfo) - 2])
-        atoms = set([distro])
-        atoms.add("%s-%s" % (distro, codename))
-        releasebits = release.split(".")
-        for i in range(len(releasebits)):
-            atoms.add("%s-%s" % (distro, ".".join(releasebits[:i + 1])))
-        if distro in ["debian", "ubuntu"]:
+        distro_id = "".join(distro_id.split()).lower()
+        atoms = set([distro_id])
+        atoms.update(self.codenamebits(distro_id, codename))
+        atoms.update(self.releasebits(distro_id, release))
+        if distro_id in ["debian", "ubuntu"]:
             atoms.add("dpkg")
+            # https://wiki.debian.org/Python
+            # https://wiki.ubuntu.com/Python
+            if (distro_id == 'ubuntu' and release_version >=
+                as_ver("16.04")) or (distro_id == 'debian' and
+                                     release_version >= as_ver("10")):
+                atoms.add(PY3)
+            else:
+                atoms.add(PY2)
             self.platform = Dpkg()
-        elif distro in ["amazonami", "centos", "redhatenterpriseserver",
-                        "redhatenterpriseworkstation",
-                        "fedora", "opensuseproject", "opensuse",
-                        "suselinux"]:
-            if distro in ["redhatenterpriseserver",
-                          "redhatenterpriseworkstation"]:
+        # RPM distros seem to be especially complicated
+        elif distro_id in ["amzn", "amazonami",
+                           "centos", "rhel",
+                           "redhatenterpriseserver",
+                           "redhatenterpriseworkstation",
+                           "fedora",
+                           "opensuseproject", "opensuse", "opensuse-leap",
+                           "opensuse-tumbleweed", "sles", "suselinux",
+                           "ol", "oracle",
+                           "rocky", "almalinux"]:
+            # Distro aliases
+            if distro_id in ["redhatenterpriseserver",
+                             "redhatenterpriseworkstation"]:
                 # just short alias
                 atoms.add("rhel")
-            elif distro == "opensuseproject":
+                atoms.update(self.codenamebits("rhel", codename))
+                atoms.update(self.releasebits("rhel", release))
+                atoms.add(PY2 if release_version < as_ver("8") else PY3)
+            elif distro_id == 'rhel' and 'server' in distro.name().lower():
+                atoms.add("redhatenterpriseserver")
+                atoms.update(self.codenamebits("redhatenterpriseserver",
+                                               codename))
+                atoms.update(self.releasebits("redhatenterpriseserver",
+                                              release))
+                atoms.add(PY2 if release_version < as_ver("8") else PY3)
+            elif (distro_id == 'rhel' and
+                    'workstation' in distro.name().lower()):
+                atoms.add("redhatenterpriseworkstation")
+                atoms.update(self.codenamebits("redhatenterpriseworkstation",
+                                               codename))
+                atoms.update(self.releasebits("redhatenterpriseworkstation",
+                                              release))
+                atoms.add(PY2 if release_version < as_ver("8") else PY3)
+            elif "amzn" in distro_id:
+                atoms.add("amazonami")
+                atoms.update(self.codenamebits("amazonami", codename))
+                atoms.update(self.releasebits("amazonami", release))
+            elif "amazonami" in distro_id:
+                atoms.add("amzn")
+                atoms.update(self.codenamebits("amzn", codename))
+                atoms.update(self.releasebits("amzn", release))
+            elif "opensuse" in distro_id:
                 # just short alias
                 atoms.add("opensuse")
+                atoms.update(self.codenamebits("opensuse", codename))
+                atoms.update(self.releasebits("opensuse", release))
+                atoms.add("opensuseproject")
+                atoms.update(self.codenamebits("opensuseproject", codename))
+                atoms.update(self.releasebits("opensuseproject", release))
+            elif "sles" in distro_id:
+                atoms.add("suselinux")
+                atoms.update(self.codenamebits("suselinux", codename))
+                atoms.update(self.releasebits("suselinux", release))
+            elif "suselinux" in distro_id:
+                atoms.add("sles")
+                atoms.update(self.codenamebits("sles", codename))
+                atoms.update(self.releasebits("sles", release))
+            elif "fedora" in distro_id:
+                atoms.add(PY2 if release_version < as_ver("23") else PY3)
+            elif any(w in distro_id for w in ["centos", "oracle", "rhel",
+                                              "rocky", "almalinux"]):
+                atoms.add(PY2 if release_version < as_ver("8") else PY3)
+
             # Family aliases
-            if 'suse' in distro:
+            if 'suse' in distro_id or distro_id == 'sles':
                 atoms.add("suse")
             else:
                 atoms.add("redhat")
 
             atoms.add("rpm")
             self.platform = Rpm()
-        elif distro in ["gentoo"]:
+        elif distro_id in ["gentoo"]:
             atoms.add("emerge")
             self.platform = Emerge()
-        elif distro in ["arch"]:
+        elif distro_id in ["arch", "manjaro"]:
             atoms.add("pacman")
             self.platform = Pacman()
+        elif distro_id in ["alpine"]:
+            atoms.add("apk")
+            self.platform = Apk()
+        else:
+            self.platform = Unknown()
         return ["platform:%s" % (atom,) for atom in sorted(atoms)]
 
 
@@ -344,6 +440,13 @@ class Platform(object):
         raise NotImplementedError(self.get_pkg_version)
 
 
+class Unknown(Platform):
+    """Unknown platform implementation. Raises error."""
+
+    def get_pkg_version(self, pkg_name):
+        raise Exception("Uknown package manager for current platform.")
+
+
 class Brew(Platform):
     """brew specific platform implementation."""
 
@@ -435,6 +538,10 @@ class Emerge(Platform):
                 ['equery', 'l', '--format=\'$version\'', pkg_name],
                 stderr=subprocess.STDOUT).decode(getpreferredencoding(False))
         except subprocess.CalledProcessError as e:
+            # The output is currently not used in the test cases.
+            # Nevertheless the next statement checks if the given
+            # parameter has the correct type.
+            e.output.decode(getpreferredencoding(False))
             if e.returncode == 3:
                 return None
             raise
@@ -458,7 +565,8 @@ class Pacman(Platform):
                 stderr=subprocess.STDOUT).decode(getpreferredencoding(False))
         except subprocess.CalledProcessError as e:
             eoutput = e.output.decode(getpreferredencoding(False))
-            if e.returncode == 1 and eoutput.endswith('was not found'):
+            s = "error: package '{}' was not found".format(pkg_name)
+            if e.returncode == 1 and s in eoutput:
                 return None
             raise
         # output looks like
@@ -467,6 +575,28 @@ class Pacman(Platform):
         return elements[1]
 
 
+class Apk(Platform):
+    """apk (Alpine Linux) specific implementation.
+
+    This shells out to apk
+    """
+
+    def get_pkg_version(self, pkg_name):
+        apk = subprocess.Popen(
+            ['apk', 'version', pkg_name], stdout=subprocess.PIPE)
+        output, _ = apk.communicate()
+        if apk.returncode == 1:
+            return None
+        output = output.decode(getpreferredencoding(False))
+        # output looks like
+        # version
+        output = output.strip()
+        elements = output.split()
+        if len(elements) < 4:
+            return None
+        return elements[4]
+
+
 def _eval_diff(operator, diff):
     """Return the boolean result for operator given diff.
 
@@ -562,8 +692,8 @@ def _cmp_segment(l_str, r_str):
     while (r_offset < len(r_str)) or (l_offset < len(l_str)):
         r_char = r_str[r_offset:r_offset + 1]
         l_char = l_str[l_offset:l_offset + 1]
-        if ((not r_char or r_char.isdigit())
-            and (not l_char or l_char.isdigit())):
+        if ((not r_char or r_char.isdigit()) and
+            (not l_char or l_char.isdigit())):
             l_int, l_offset = _find_int(l_str, l_offset)
             r_int, r_offset = _find_int(r_str, r_offset)
             diff = l_int - r_int
