@@ -16,39 +16,83 @@
 # limitations under the License.
 
 import contextlib
+import os.path
 import platform
 import subprocess
 from textwrap import dedent
 
+import distro
 import fixtures
 import mock
 import ometa.runtime
+from testtools import ExpectedException
 from testtools.matchers import Contains
 from testtools.matchers import Equals
 from testtools.matchers import MatchesSetwise
 from testtools import TestCase
 
 from bindep.depends import _eval
+from bindep.depends import Brew
 from bindep.depends import Depends
 from bindep.depends import Dpkg
 from bindep.depends import Emerge
 from bindep.depends import Pacman
 from bindep.depends import Rpm
+from bindep.depends import Apk
 
 
 # NOTE(notmorgan): In python3 subprocess.check_output returns bytes not
 # string. All mock calls for subprocess.check_output have been updated to
 # ensure bytes is used over string. In python 2 this is a no-op change.
 
+FIXTURE_DIR = os.path.join(os.path.dirname(__file__),
+                           'fixtures')
+
+
+class DistroFixture(fixtures.Fixture):
+    def __init__(self, distro_name):
+        self.distro_name = distro_name.lower()
+
+    def _setUp(self):
+        # Python 2.7 and 3.5 need older distro which has different src paths
+        if distro.__version__ < '1.7.0':
+            distro_patch_path = 'distro._distro'
+        else:
+            distro_patch_path = 'distro.distro._distro'
+        # This type of monkey patching is borrowed from the distro test
+        # suite.
+        os_release = os.path.join(FIXTURE_DIR, self.distro_name,
+                                  'etc', 'os-release')
+        mydistro = distro.LinuxDistribution(False, os_release, 'non')
+        self.useFixture(
+            fixtures.MonkeyPatch(distro_patch_path, mydistro))
+        if self.distro_name not in ['darwin']:
+            self.useFixture(fixtures.MonkeyPatch(
+                'platform.system', lambda *x: 'Linux'))
+        else:
+            self.useFixture(fixtures.MonkeyPatch(
+                'distro.id', lambda *x: self.distro_name))
+
 
 class TestDepends(TestCase):
 
+    def test_filename(self):
+        depends = Depends("")
+        self.assertIsNone(depends.filename)
+        depends = Depends("", "/some/path/to/bindep.txt")
+        self.assertEqual(depends.filename, "/some/path/to/bindep.txt")
+
     def test_empty_file(self):
         depends = Depends("")
         self.assertEqual([], depends.profiles())
 
+    def test_3tuple(self):
+        depends = Depends(u"erlang [(infra rabbitmq hipe)]\n")
+        self.assertEqual(sorted([u'infra', u'rabbitmq', u'hipe']),
+                         depends.profiles())
+
     def test_platform_profiles_succeeds(self):
-        with self._mock_lsb('Ubuntu'):
+        with DistroFixture('Ubuntu'):
             depends = Depends("")
             self.assertIsInstance(depends.platform_profiles(), list)
 
@@ -75,14 +119,24 @@ class TestDepends(TestCase):
         yield mock_checkoutput
         mock_checkoutput.assert_called_once_with()
 
+    def test_detects_unknown(self):
+        with DistroFixture("Unknown"):
+            depends = Depends("")
+            self.assertThat(
+                depends.platform_profiles(), Contains("platform:unknown"))
+            with ExpectedException(Exception,
+                                   "Uknown package manager for "
+                                   "current platform."):
+                depends.platform.get_pkg_version('x')
+
     def test_detects_amazon_linux(self):
-        with self._mock_lsb("AmazonAMI"):
+        with DistroFixture("AmazonAMI"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:amazonami"))
 
     def test_detects_centos(self):
-        with self._mock_lsb("CentOS"):
+        with DistroFixture("CentOS"):
             depends = Depends("")
             platform_profiles = depends.platform_profiles()
             self.assertThat(platform_profiles, Contains("platform:centos"))
@@ -94,8 +148,8 @@ class TestDepends(TestCase):
             platform_profiles = depends.platform_profiles()
             self.assertThat(platform_profiles, Contains("platform:darwin"))
 
-    def test_detects_rhel(self):
-        with self._mock_lsb("RedHatEnterpriseServer"):
+    def test_detects_rhel7(self):
+        with DistroFixture("RHELServer7"):
             depends = Depends("")
             platform_profiles = depends.platform_profiles()
             self.assertThat(
@@ -107,9 +161,10 @@ class TestDepends(TestCase):
             self.assertThat(
                 platform_profiles,
                 Contains("platform:redhat"))
+            self.assertThat(platform_profiles, Contains("platform:base-py2"))
 
     def test_detects_rhel_workstation(self):
-        with self._mock_lsb("RedHatEnterpriseWorkstation"):
+        with DistroFixture("RHELWorkstation7"):
             depends = Depends("")
             platform_profiles = depends.platform_profiles()
             self.assertThat(
@@ -121,16 +176,27 @@ class TestDepends(TestCase):
             self.assertThat(
                 platform_profiles,
                 Contains("platform:redhat"))
+            self.assertThat(platform_profiles, Contains("platform:base-py2"))
 
-    def test_detects_fedora(self):
-        with self._mock_lsb("Fedora"):
+    def test_detects_fedora23(self):
+        with DistroFixture("Fedora23"):
             depends = Depends("")
             platform_profiles = depends.platform_profiles()
             self.assertThat(platform_profiles, Contains("platform:fedora"))
             self.assertThat(platform_profiles, Contains("platform:redhat"))
+            self.assertThat(platform_profiles, Contains("platform:base-py3"))
+
+    def test_detects_rhel8(self):
+        with DistroFixture("rhel8"):
+            depends = Depends("")
+            platform_profiles = depends.platform_profiles()
+            self.assertThat(platform_profiles, Contains("platform:redhat"))
+            self.assertThat(platform_profiles, Contains("platform:base-py3"))
 
     def test_detects_opensuse_project(self):
-        with self._mock_lsb("openSUSE Project"):
+        # TODO what does an os-release for opensuse project look like?
+        # Is this different than sles, leap, and tumbleweed?
+        with DistroFixture("openSUSEleap"):
             depends = Depends("")
             platform_profiles = depends.platform_profiles()
             self.assertThat(platform_profiles,
@@ -138,12 +204,21 @@ class TestDepends(TestCase):
             self.assertThat(platform_profiles,
                             Contains("platform:opensuse"))
             self.assertThat(platform_profiles,
-                            Contains("platform:opensuseproject-14.04"))
+                            Contains("platform:opensuseproject-42.1"))
             self.assertThat(platform_profiles,
                             Contains("platform:suse"))
 
     def test_detects_opensuse(self):
-        with self._mock_lsb("openSUSE"):
+        with DistroFixture("openSUSEleap"):
+            depends = Depends("")
+            platform_profiles = depends.platform_profiles()
+            self.assertThat(platform_profiles,
+                            Contains("platform:opensuse"))
+            self.assertThat(platform_profiles,
+                            Contains("platform:suse"))
+
+    def test_detects_opensuse_leap15(self):
+        with DistroFixture("openSUSEleap15"):
             depends = Depends("")
             platform_profiles = depends.platform_profiles()
             self.assertThat(platform_profiles,
@@ -152,98 +227,151 @@ class TestDepends(TestCase):
                             Contains("platform:suse"))
 
     def test_detects_suse_linux(self):
-        with self._mock_lsb("SUSE Linux"):
+        with DistroFixture("SLES"):
             depends = Depends("")
             platform_profiles = depends.platform_profiles()
             self.assertThat(platform_profiles, Contains("platform:suselinux"))
             self.assertThat(platform_profiles, Contains("platform:suse"))
 
     def test_detects_ubuntu(self):
-        with self._mock_lsb("Ubuntu"):
+        with DistroFixture("Ubuntu"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:ubuntu"))
 
-    def test_detects_release(self):
-        with self._mock_lsb("Ubuntu"):
+    def test_detects_alpine(self):
+        with DistroFixture("Alpine"):
             depends = Depends("")
             self.assertThat(
-                depends.platform_profiles(), Contains("platform:ubuntu-14"))
+                depends.platform_profiles(), Contains("platform:alpine"))
+
+    def test_detects_oraclelinux(self):
+        with DistroFixture("oraclelinux"):
+            depends = Depends("")
+            self.assertThat(
+                depends.platform_profiles(), Contains("platform:rpm"))
+
+    def test_detects_rocky(self):
+        with DistroFixture("rocky"):
+            depends = Depends("")
+            self.assertThat(
+                depends.platform_profiles(), Contains("platform:rpm"))
+
+    def test_detects_almalinux(self):
+        with DistroFixture("almalinux"):
+            depends = Depends("")
+            self.assertThat(
+                depends.platform_profiles(), Contains("platform:rpm"))
+
+    def test_detects_release(self):
+        with DistroFixture("Ubuntu"):
+            depends = Depends("")
+            self.assertThat(
+                depends.platform_profiles(), Contains("platform:ubuntu-16"))
 
     def test_detects_subrelease(self):
-        with self._mock_lsb("Ubuntu"):
+        with DistroFixture("Ubuntu"):
             depends = Depends("")
             self.assertThat(
-                depends.platform_profiles(), Contains("platform:ubuntu-14.04"))
+                depends.platform_profiles(), Contains("platform:ubuntu-16.04"))
 
     def test_detects_codename(self):
-        with self._mock_lsb("Ubuntu"):
+        with DistroFixture("Ubuntu"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(),
-                Contains("platform:ubuntu-trusty"))
+                Contains("platform:ubuntu-xenial"))
 
     def test_centos_implies_rpm(self):
-        with self._mock_lsb("CentOS"):
+        with DistroFixture("CentOS"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:rpm"))
             self.assertIsInstance(depends.platform, Rpm)
 
+    def test_darwin_implies_brew(self):
+        with self._mock_platform_darwin("Darwin"):
+            with DistroFixture("Darwin"):
+                # to make it believe `which brew` succeeded
+                self.useFixture(
+                    fixtures.MockPatchObject(
+                        os,
+                        'system',
+                        return_value=0)).mock
+                depends = Depends("")
+                self.assertThat(
+                    depends.platform_profiles(), Contains("platform:brew"))
+                self.assertIsInstance(depends.platform, Brew)
+
     def test_rhel_implies_rpm(self):
-        with self._mock_lsb("RedHatEnterpriseServer"):
+        with DistroFixture("RHELServer7"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:rpm"))
             self.assertIsInstance(depends.platform, Rpm)
 
     def test_fedora_implies_rpm(self):
-        with self._mock_lsb("Fedora"):
+        with DistroFixture("Fedora23"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:rpm"))
             self.assertIsInstance(depends.platform, Rpm)
 
     def test_opensuse_project_implies_rpm(self):
-        with self._mock_lsb("openSUSE Project"):
+        with DistroFixture("openSUSEleap"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:rpm"))
             self.assertIsInstance(depends.platform, Rpm)
 
     def test_opensuse_implies_rpm(self):
-        with self._mock_lsb("openSUSE"):
+        with DistroFixture("openSUSEleap"):
+            depends = Depends("")
+            self.assertThat(
+                depends.platform_profiles(), Contains("platform:rpm"))
+            self.assertIsInstance(depends.platform, Rpm)
+
+    def test_opensuse_leap15_implies_rpm(self):
+        with DistroFixture("openSUSEleap15"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:rpm"))
             self.assertIsInstance(depends.platform, Rpm)
 
     def test_suse_linux_implies_rpm(self):
-        with self._mock_lsb("SUSE LINUX"):
+        with DistroFixture("SLES"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:rpm"))
             self.assertIsInstance(depends.platform, Rpm)
 
     def test_ubuntu_implies_dpkg(self):
-        with self._mock_lsb("Ubuntu"):
+        with DistroFixture("Ubuntu"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:dpkg"))
             self.assertIsInstance(depends.platform, Dpkg)
 
+    def test_alpine_implies_apk(self):
+        with DistroFixture("Alpine"):
+            depends = Depends("")
+            self.assertThat(
+                depends.platform_profiles(), Contains("platform:apk"))
+            self.assertIsInstance(depends.platform, Apk)
+
     def test_arch_implies_pacman(self):
-        with self._mock_lsb("Arch"):
+        with DistroFixture("Arch"):
             depends = Depends("")
             self.assertThat(
                 depends.platform_profiles(), Contains("platform:pacman"))
             self.assertIsInstance(depends.platform, Pacman)
 
-    def test_missing_lsb_release(self):
-        with mock.patch('subprocess.check_output') as mock_co:
-            mock_co.side_effect = OSError
+    def test_manjaro_implies_pacman(self):
+        with DistroFixture("manjaro"):
             depends = Depends("")
-            self.assertRaises(OSError, depends.platform_profiles)
+            self.assertThat(
+                depends.platform_profiles(), Contains("platform:pacman"))
+            self.assertIsInstance(depends.platform, Pacman)
 
     def test_finds_profiles(self):
         depends = Depends(dedent("""\
@@ -260,6 +388,12 @@ class TestDepends(TestCase):
         depends = Depends("")
         self.assertEqual([], depends._rules)
 
+    def test_no_newline(self):
+        depends = Depends("foo")
+        self.assertEqual(
+            [("foo", [], [])],
+            depends._rules)
+
     def test_selectors(self):
         depends = Depends("foo [!bar baz quux]\n")
         self.assertEqual(
@@ -388,6 +522,15 @@ class TestDepends(TestCase):
             """))
         self.assertEqual(len(depends.active_rules(['default'])), 3)
 
+    def test_parser_accepts_full_path_to_tools(self):
+        # at least yum/dnf allow these instead of mentioning rpm names
+        depends = Depends(dedent("""\
+            /usr/bin/bash
+            """))
+        self.assertEqual(
+            [("/usr/bin/bash", [], [])],
+            depends.active_rules(["default"]))
+
     def test_parser_invalid(self):
         self.assertRaises(ometa.runtime.ParseError,
                           lambda: Depends("foo [platform:bar@baz]\n"))
@@ -586,7 +729,7 @@ class TestEmerge(TestCase):
         platform = Emerge()
 
         def _side_effect_raise(*args, **kwargs):
-            raise subprocess.CalledProcessError(3, [], '')
+            raise subprocess.CalledProcessError(3, [], b'')
 
         mocked_checkoutput = self.useFixture(
             fixtures.MockPatchObject(subprocess, 'check_output')).mock
@@ -601,7 +744,7 @@ class TestEmerge(TestCase):
         platform = Emerge()
 
         def _side_effect_raise(*args, **kwargs):
-            raise subprocess.CalledProcessError(3, [], '')
+            raise subprocess.CalledProcessError(3, [], b'')
 
         mocked_checkoutput = self.useFixture(
             fixtures.MockPatchObject(subprocess, 'check_output')).mock
@@ -653,6 +796,47 @@ class TestPacman(TestCase):
             stderr=subprocess.STDOUT)
 
 
+class TestApk(TestCase):
+
+    @mock.patch('subprocess.Popen')
+    def test_unknown_package(self, mock_popen):
+        platform = Apk()
+
+        proc_mock = mock.Mock()
+        proc_mock.communicate.return_value = (b'Installed: Available:\n', b'')
+        mock_popen.returncode = 1
+        mock_popen.return_value = proc_mock
+
+        self.assertEqual(None, platform.get_pkg_version("foo"))
+        mock_popen.assert_called_once_with(
+            ['apk', 'version', 'foo'], stdout=subprocess.PIPE)
+        self.assertEqual(None, platform.get_pkg_version("foo"))
+
+    @mock.patch('subprocess.Popen')
+    def test_installed_version(self, mock_popen):
+        platform = Apk()
+
+        proc_mock = mock.Mock()
+        proc_mock.communicate.return_value = (
+            b'Insd: Able: foo-4.0.0-r1 = 4.0.0-r1', b'')
+        mock_popen.return_value = proc_mock
+        self.assertEqual('4.0.0-r1', platform.get_pkg_version("foo"))
+        mock_popen.assert_called_once_with(
+            ['apk', 'version', 'foo'], stdout=subprocess.PIPE)
+
+    @mock.patch('subprocess.Popen')
+    def test_handles_warnings(self, mock_popen):
+        platform = Apk()
+        process_mock = mock.Mock()
+        values = b'Installed:   Available:\n'
+        warnings = b'WARNING: Ignoring APKINDEX.blah\nWARNING2: blah blah\n'
+        process_mock.communicate.return_value = (values, warnings)
+        mock_popen.return_value = process_mock
+        self.assertIsNone(platform.get_pkg_version("bar"))
+        mock_popen.assert_called_once_with(
+            ['apk', 'version', 'bar'], stdout=subprocess.PIPE)
+
+
 class TestRpm(TestCase):
 
     # NOTE: test_not_installed is not implemented as rpm seems to only be aware
